// ============================================================================
// CBP + FBO = MAS22 Simulation Framework
// 20-12-21
//
// ============================================================================

#pragma once

#include "CoreMinimal.h"
#include "GameFramework/Actor.h"
#include "MAS22_Structs.h"
#include "MAS22_Actor.generated.h"

// (12) -----------------------
UENUM()
enum class SolverEnum : uint8 {
	EULER					UMETA(DisplayName = "Euler"),
	EULER_CROMER			UMETA(DisplayName = "Euler-Cromer"),
	RUNGE_KUTTA				UMETA(DisplayName = "Runge-Kutta"),
	RUNGE_KUTTA_FEHLBERG	UMETA(DisplayName = "Runge-Kutta-Fehlberg"),
};

UCLASS()
class MAS22_SCIENCEPARK_API AMAS22_Actor : public AActor
{
	GENERATED_BODY()
	
public:	
	AMAS22_Actor();

	virtual void ActorInitialize();
	virtual void SetRunning();

	virtual void SetInitialConditions();
	// (12) ------------------------------

	virtual void SetParams(double paramArray[]);

	virtual void Message();

	TArray<HUDLine> HUDLines;
	void GetHUDLines(TArray<HUDLine>** plines);

	int ID = 2;

	bool bRunning = false;

	bool bHUD_exists = false;

	FTimerHandle TimerHandle;

	// CBP 20-12-21 Name added. Used in HUD and to create unique log file
	UPROPERTY(EditAnywhere, Category = "MAS22 Simulation Engine")
		FString actorName;

	// (12) -------------------------
	UPROPERTY(EditAnywhere, Category = "MAS22 Simulation Engine")
		SolverEnum solver;
	UPROPERTY(EditAnywhere, Category = "MAS22 Simulation Engine")
		double dT = 0.01;

	// (18) --------------------------
	UPROPERTY(EditAnywhere, Category = "MAS22 Simulation Engine")
		double logInterval = 0.1;

protected:
	virtual void LogData();
	virtual void EndPlay(const EEndPlayReason::Type EndPlayReason) override;

	virtual void ComputeDT();
	virtual void Computation();
	virtual void Visualization(float deltaTime);
	virtual void BuildHUDLines();


	double time;
	int itn;

public:	
	
	//(12) -------------------------
	virtual void InitializeSolver();
	void Euler(double DeltaTime);
	void EulerCromer(double DeltaTime);
	void RungeKutta(double DeltaTime);
	void RungeKuttaFehlberg(double DeltaTime);
	//void SetParams(double paramArray[]);

	// (12) -------------------------
	// for RKF -----------
	double *u;
	double *dudt;
	double t;
	double tout;
	double relerr;
	double abserr;
	int flag;	

	// (20) For implementation of solver RHS
	// params and options.
	double* parameters;
	// (21) options removed
	//double* options;

};
