// ============================================================================
// CBP + FBO = MAS22 Simulation Framework
// 20-12-21
// 28-01-22
// MAS22_Monster_Truck.cpp
// ============================================================================

#include "MAS22_Monster_Truck.h"
#include "MAS22_Actor.h"
#include "MAS22_Structs.h"
#include "rkf45.hpp"
#include "rhs_Monster_Truck.h"

//(11) ----------------------------
#include "Components/CapsuleComponent.h"
#include "Components/StaticMeshComponent.h"

AMAS22_Monster_Truck::AMAS22_Monster_Truck() {

	PrimaryActorTick.bCanEverTick = false;

	// (11) Add components
	CapsuleCollider = CreateDefaultSubobject<UCapsuleComponent>("CapsuleCollider");
	CapsuleCollider->SetupAttachment(GetRootComponent());
	CapsuleCollider->SetCapsuleHalfHeight(200);
	CapsuleCollider->SetCapsuleRadius(700);
	CapsuleCollider->SetRelativeLocation(FVector(0, 0, 330));

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset1(TEXT("/Game/Car/MTruckBody.MTruckBody"));
	UStaticMesh* Asset1 = MeshAsset1.Object;

	Body = CreateDefaultSubobject<UStaticMeshComponent>("Body");
	Body->SetStaticMesh(Asset1);
	Body->SetRelativeLocation(FVector(0, 0, 330));
	Body->SetupAttachment(CapsuleCollider);

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset2(TEXT("/Game/Car/MTruckAxle.MTruckAxle"));
	UStaticMesh *Asset2 = MeshAsset2.Object;

	Axle = CreateDefaultSubobject<UStaticMeshComponent>("Axle");
	Axle->SetStaticMesh(Asset2);
	Axle->SetRelativeLocation(FVector(20, 0, 100));
	Axle->SetupAttachment(CapsuleCollider);

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset3(TEXT("/Game/Car/MTruckWheel.MTruckWheel"));
	UStaticMesh* Asset3 = MeshAsset3.Object;

	Wheel_LF = CreateDefaultSubobject<UStaticMeshComponent>("Wheel_LF");
	Wheel_LF->SetStaticMesh(Asset3);
	Wheel_LF->SetRelativeLocation(FVector(-180, 220, 100));
	Wheel_LF->SetupAttachment(CapsuleCollider);

	Wheel_RF = CreateDefaultSubobject<UStaticMeshComponent>("Wheel_RF");
	Wheel_RF->SetStaticMesh(Asset3);
	Wheel_RF->SetRelativeLocation(FVector(-180, -150, 100));
	Wheel_RF->SetupAttachment(CapsuleCollider);

	Wheel_LR = CreateDefaultSubobject<UStaticMeshComponent>("Wheel_LR");
	Wheel_LR->SetStaticMesh(Asset3);
	Wheel_LR->SetRelativeLocation(FVector(220, 220, 100));
	Wheel_LR->SetupAttachment(CapsuleCollider);

	Wheel_RR = CreateDefaultSubobject<UStaticMeshComponent>("Wheel_RR");
	Wheel_RR->SetStaticMesh(Asset3);
	Wheel_RR->SetRelativeLocation(FVector(220, -150, 100));
	Wheel_RR->SetupAttachment(CapsuleCollider);

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset4(TEXT("/Game/Car/MTruckSpring.MTruckSpring"));
	UStaticMesh* Asset4 = MeshAsset4.Object;

	Spring_LF = CreateDefaultSubobject<UStaticMeshComponent>("Spring_LF");
	Spring_LF->SetStaticMesh(Asset4);
	Spring_LF->SetRelativeLocation(FVector(-180, 100, 180));
	Spring_LF->SetupAttachment(CapsuleCollider);

	Spring_RF = CreateDefaultSubobject<UStaticMeshComponent>("Spring_RF");
	Spring_RF->SetStaticMesh(Asset4);
	Spring_RF->SetRelativeLocation(FVector(-180, -100, 180));
	Spring_RF->SetupAttachment(CapsuleCollider);

	Spring_LR = CreateDefaultSubobject<UStaticMeshComponent>("Spring_LR");
	Spring_LR->SetStaticMesh(Asset4);
	Spring_LR->SetRelativeLocation(FVector(220, 100, 180));
	Spring_LR->SetupAttachment(CapsuleCollider);

	Spring_RR = CreateDefaultSubobject<UStaticMeshComponent>("Spring_RR");
	Spring_RR->SetStaticMesh(Asset4);
	Spring_RR->SetRelativeLocation(FVector(220, -100, 180));
	Spring_RR->SetupAttachment(CapsuleCollider);

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset5(TEXT("/Game/Car/monsterTruckPlate.monsterTruckPlate"));
	UStaticMesh* Asset5 = MeshAsset5.Object;

	FloorPlate = CreateDefaultSubobject<UStaticMeshComponent>("FloorPlate");
	FloorPlate->SetStaticMesh(Asset5);
	FloorPlate->SetRelativeLocation(FVector(0, 0, 0));
	FloorPlate->SetupAttachment(CapsuleCollider);

	floorPlateStartLocation = FloorPlate->GetRelativeLocation();
	bodyStartLocation = Body->GetRelativeLocation();
	axleStartLocation = Axle->GetRelativeLocation();
	Wheel_LF_StartLocation = Wheel_LF->GetRelativeLocation();
	Wheel_LR_StartLocation = Wheel_LR->GetRelativeLocation();
	Wheel_RF_StartLocation = Wheel_RF->GetRelativeLocation();
	Wheel_RR_StartLocation = Wheel_RR->GetRelativeLocation();
	Spring_LF_StartLocation = Spring_LF->GetRelativeLocation();
	Spring_LR_StartLocation = Spring_LR->GetRelativeLocation();
	Spring_RF_StartLocation = Spring_RF->GetRelativeLocation();
	Spring_RR_StartLocation = Spring_RR->GetRelativeLocation();

}

// Called when the game starts or when spawned
void AMAS22_Monster_Truck::ActorInitialize() {
	Super::ActorInitialize();
	// (13) Allocate arrays - note these are pointers
	u = new double[2];
	dudt = new double[2];
	// (20) ----------- params
	parameters = new double[6];
	// (20) ----------- options
	// (21) ---------removed
	// options = new double[1];

	// (15) ----------------
	SetInitialConditions();
	// (20) ----------------
	InitializeSolver();
}

void AMAS22_Monster_Truck::BuildHUDLines() {
	HUDLine line;
	Super::BuildHUDLines();
	HUDLines.Empty();
	bHUD_exists = true;
	line.message = FString::Printf(TEXT("time = %f"), time);
	line.colour = FColor::Red;
	line.value = 0;
	HUDLines.Add(line);
	line.message = FString::Printf(TEXT("floor z = %f"), floorPlateDispZ);
	line.colour = FColor::Yellow;
	line.value = 0;
	HUDLines.Add(line);
	line.message = FString::Printf(TEXT("u[0] = %f"), u[0]);
	line.colour = FColor::Yellow;
	line.value = 0;
	HUDLines.Add(line);
	line.message = FString::Printf(TEXT("maxDispZ = %f"), maxUO);
	line.colour = FColor::Yellow;
	line.value = 0;
	HUDLines.Add(line);
	line.message = FString::Printf(TEXT("accelZ = %f"), dudt[1]);
	line.colour = FColor::Yellow;
	line.value = 0;
	HUDLines.Add(line);
	line.message = FString::Printf(TEXT("maxAccelZ (g's) = %f"), maxAccelZ/9.8);
	line.colour = FColor::Yellow;
	line.value = 0;
	HUDLines.Add(line);
}

void AMAS22_Monster_Truck::SetInitialConditions() {
	Super::SetInitialConditions();
	// (15) ----------------------------
	u[0] = initDispZ;
	u[1] = 0.0;
	
}

void AMAS22_Monster_Truck::InitializeSolver() {
	// (20) set parameters array here - just for the moment (30-01-22)
	// (22) -----------------
	if (selectedRHS == coil) {
		parameters[0] = mass;
		parameters[1] = springConstant;
		parameters[2] = damping;
	}
	if (selectedRHS == air) {
		parameters[0] = mass;
		//parameters[1] = springConstant;
		parameters[2] = damping;
		parameters[3] = air_A;
		parameters[4] = air_x0;
		parameters[5] = air_k;
	}

	setParams(parameters);

	// (21) -----------------------------
	// RKF Stuff
	flag = 1;
	t = 0.0;
	relerr = 100 * sqrt(r8_epsilon());
	abserr = 100 * sqrt(r8_epsilon());
	GEngine->AddOnScreenDebugMessage(-1, 10.0f, FColor::Yellow,
			FString::Printf(TEXT("MAS22_MonsterTruck:: relerr  = %e"),relerr));

}

// (19) ---------------------------------
void AMAS22_Monster_Truck::LogData() {
	int m;
	m = (int(logInterval / dT));
	if (itn % m == 0) {
		dataRecords.Add(FDataRecord{ time, u[0], u[1] });
	}

}

// (12) -------------------------------------
void AMAS22_Monster_Truck::ComputeDT() {
	if (bRunning) {
		Computation();
		Visualization(dT);
		BuildHUDLines();
		LogData();
		if (u[0] > maxUO) {
			maxUO = u[0];
		}
		if (u[0] < 0.1) {	// Arbitrary - need to revisit
			maxUO = 0;
		}
		if (abs(dudt[1]) > maxAccelZ) {
			maxAccelZ = dudt[1];
		}

		itn++;
		//GEngine->AddOnScreenDebugMessage(-1, 4.0f, FColor::Yellow,
		//	FString::Printf(TEXT("MAS22_MonsterTruck:: disp = %f"),u[0]));
	}
}

void AMAS22_Monster_Truck::Computation() {
	// (21) --------------------
	switch (forcingType) {
	case step:
		floorPlateDispZ = stepHeight;
		applyFloorPlateDispZ(floorPlateDispZ);
		applyStaticForce(0.0);
		break;
	case periodic:
		floorPlateDispZ = forcingAmplitude * FMath::Cos(2 * PI * time / forcingPeriod);
		applyFloorPlateDispZ(floorPlateDispZ);
		applyStaticForce(0.0);
		break;
	}
	 

	switch(solver) {
	case SolverEnum::EULER:
		Euler(dT);
		break;
		// (21) -----------------------------
	case SolverEnum::RUNGE_KUTTA_FEHLBERG:
		RKF(dT);
		break;
	default :
		break;
	}

	// (21) move ----------------------------
	//time += dT;
}

// (15) ------------------------------------

void AMAS22_Monster_Truck::Euler(double deltaT) {
	
	dT = deltaT;
	// (22) ---------------------------
	if (selectedRHS == coil) {
		rhs_Monster_Truck(time, u, dudt);
	}
	if (selectedRHS == air) {
		rhs_Monster_Truck_Air(time, u, dudt);
	}

	u[0] += dudt[0] * dT;
	u[1] += dudt[1] * dT;
	time += dT;

	
}

//  (21) -------------------------------------
void AMAS22_Monster_Truck::RKF(double deltaT) {
	flag = 2;
	tout = time + deltaT; 
	if (selectedRHS == coil) {
		flag = r8_rkf45(rhs_Monster_Truck, 2, u, dudt, &time, tout, &relerr, abserr, flag);
	}
	if (selectedRHS == air) {
		flag = r8_rkf45(rhs_Monster_Truck_Air, 2, u, dudt, &time, tout, &relerr, abserr, flag);
	}

	//GEngine->AddOnScreenDebugMessage(-1, 4.0f, FColor::Yellow,
			//FString::Printf(TEXT("MAS22_MonsterTruck::RKF() flag = %d"),flag));
	//if (flag == 2) flag = -2;


}

void AMAS22_Monster_Truck::Visualization(float deltaTime) {
	// (16) ---------------------------------------
	double springLength = 142;
	double dispZ;
	dispZ = u[0];

	// Move the FloorPlate
	FVector newLocation = floorPlateStartLocation;
	newLocation.Z += scaling * floorPlateDispZ;
	FloorPlate->SetRelativeLocation(newLocation);

	// Move the Body
	newLocation = bodyStartLocation;
	newLocation.Z += scaling*dispZ;
	Body->SetRelativeLocation(newLocation);

	// Move the Axles
	newLocation = axleStartLocation;
	newLocation.Z += scaling * floorPlateDispZ;
	Axle->SetRelativeLocation(newLocation);

	// Move the Wheels
	newLocation = Wheel_LF_StartLocation;
	newLocation.Z += scaling * floorPlateDispZ;
	Wheel_LF->SetRelativeLocation(newLocation);

	newLocation = Wheel_RF_StartLocation;
	newLocation.Z += scaling * floorPlateDispZ;
	Wheel_RF->SetRelativeLocation(newLocation);

	newLocation = Wheel_LR_StartLocation;
	newLocation.Z += scaling * floorPlateDispZ;
	Wheel_LR->SetRelativeLocation(newLocation);

	newLocation = Wheel_RR_StartLocation;
	newLocation.Z += scaling * floorPlateDispZ;
	Wheel_RR->SetRelativeLocation(newLocation);

	// Move the Springs
	newLocation = Spring_LF_StartLocation;
	newLocation.Z += scaling * (floorPlateDispZ + dispZ) / 2.0;
	Spring_LF->SetRelativeLocation(newLocation);

	
	newLocation = Spring_LR_StartLocation;
	newLocation.Z += scaling * (floorPlateDispZ + dispZ) / 2.0;
	Spring_LR->SetRelativeLocation(newLocation);

	
	newLocation = Spring_RF_StartLocation;
	newLocation.Z += scaling * (floorPlateDispZ + dispZ) / 2.0;
	Spring_RF->SetRelativeLocation(newLocation);

	newLocation = Spring_RR_StartLocation;
	newLocation.Z += scaling * (floorPlateDispZ + dispZ) / 2.0;
	Spring_RR->SetRelativeLocation(newLocation);

	// Scale the Springs
	FVector newVec;
	newVec = Spring_LF->GetRelativeScale3D();
	newVec.Z = (springLength + (dispZ - floorPlateDispZ) * scaling) / springLength;
	Spring_LF->SetRelativeScale3D(newVec);

	newVec = Spring_LR->GetRelativeScale3D();
	newVec.Z = (springLength + (dispZ - floorPlateDispZ) * scaling) / springLength;
	Spring_LR->SetRelativeScale3D(newVec);

	newVec = Spring_RF->GetRelativeScale3D();
	newVec.Z = (springLength + (dispZ - floorPlateDispZ) * scaling) / springLength;
	Spring_RF->SetRelativeScale3D(newVec);

	newVec = Spring_RR->GetRelativeScale3D();
	newVec.Z = (springLength + (dispZ - floorPlateDispZ) * scaling) / springLength;
	Spring_RR->SetRelativeScale3D(newVec);
	
	
}

void AMAS22_Monster_Truck::EndPlay(const EEndPlayReason::Type EndPlayReason) {
	Super::EndPlay(EndPlayReason);

	// (19) Dump Octave file -------------
	FString logFile = FPaths::ProjectDir();
	FString fileName = "OctaveFiles/";
	fileName.Append(actorName);
	fileName.Append(".m");
	logFile.Append(fileName);

	int nrRecords = dataRecords.Num();
	FString stringToSave = "A=[\n";

	for (int i = 0; i < nrRecords; i++) {
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].time));
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].dispZ));
		stringToSave.Append(FString::Printf(TEXT("%e\n"), dataRecords[i].velyZ));
	}

	stringToSave.Append("];\n\n");
	stringToSave.Append("time  = A(:,1);\n");
	stringToSave.Append("dispZ = A(:,2);\n");
	stringToSave.Append("velyZ = A(:,3);\n");

	stringToSave.Append("figure(1)\n");
	stringToSave.Append("subplot(2,1,1)\n");
	stringToSave.Append("plot(time,dispZ);\n");
	stringToSave.Append("title('DispZ vs. time')\n");
	stringToSave.Append("xlabel('time (secs)')\n");
	stringToSave.Append("ylabel('dispZ (m)')\n");

	stringToSave.Append("subplot(2,1,2)\n");
	stringToSave.Append("plot(time,velyZ);\n");
	stringToSave.Append("title('VelyZ vs. time')\n");
	stringToSave.Append("xlabel('time (secs)')\n");
	stringToSave.Append("ylabel('VelyZ (m/s)')\n");

	stringToSave.Append("figure(2)\n");
	stringToSave.Append("plot(dispZ,velyZ);\n");
	stringToSave.Append("title('VelyZ vs. dispZ Phase Plane')\n");
	stringToSave.Append("xlabel('distZ ()')\n");
	stringToSave.Append("ylabel('velyZ (metres/sec)')\n");

	FFileHelper::SaveStringToFile(stringToSave, *logFile);


}