// ============================================================================
// CBP + FBO = MAS22 Simulation Framework
// 21-12-21
// 28-01-22
// // 29-01-22
// MAS22_Monster_Truck.h
// ============================================================================

#pragma once

#include "CoreMinimal.h"
#include "GameFramework/Actor.h"
#include "MAS22_Structs.h"
#include "MAS22_Actor.h"
#include "MAS22_Monster_Truck.generated.h"

UENUM()
enum ForcingType {
	periodic	    UMETA(DisplayName = "Periodic"),
	step			UMETA(DisplayName = "Stepped"),
	manual			UMETA(DisplayName = "Manual"),
	none			UMETA(DisplayName = "None")
};

// (22) ---------------------------
UENUM()
enum selectRHS {
	coil		    UMETA(DisplayName = "Coil Spring"),
	air				UMETA(DisplayName = "Air Suspension")
};

// (19) ---------------------------
USTRUCT()
struct FDataRecord {
	GENERATED_BODY()
		double time, dispZ, velyZ;
};

UCLASS()
class MAS22_SCIENCEPARK_API AMAS22_Monster_Truck : public AMAS22_Actor {
	GENERATED_BODY()

	// (11) Add components --------------------------------------------------
		// 30-01-22 Changed 'Instanced' to 'VisibleDefaultsOnly'
	UPROPERTY(VisibleDefaultsOnly)
		class UCapsuleComponent* CapsuleCollider;
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* Body;
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* Axle;
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* Wheel_LF;
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* Wheel_RF;
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* Wheel_LR;
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* Wheel_RR;
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* Spring_LF;
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* Spring_RF;
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* Spring_LR;
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* Spring_RR;
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* FloorPlate;
	
public:	
	AMAS22_Monster_Truck();
	virtual void ActorInitialize() override;
	virtual void SetInitialConditions() override;
	virtual void InitializeSolver() override;		// (20) -------------------------
	virtual void ComputeDT() override;

	// (15) ------------------------------
public:
	UPROPERTY(EditAnywhere)
		double mass;
	UPROPERTY(EditAnywhere)
		double springConstant;
	UPROPERTY(EditAnywhere)
		double damping;
	UPROPERTY(EditAnywhere)
		double scaling = 5.0;
	// (22) ---------------------------
	UPROPERTY(EditAnywhere)
		TEnumAsByte<selectRHS> selectedRHS;
	UPROPERTY(EditAnywhere)
		double air_A = 0.002;
	UPROPERTY(EditAnywhere)
		double air_x0 = 0.51;
	UPROPERTY(EditAnywhere)
		double air_k = 350;

	UPROPERTY(EditAnywhere)
		double forcingAmplitude;
	UPROPERTY(EditAnywhere)
		double forcingPeriod;
	UPROPERTY(EditAnywhere)
		double stepHeight;	
	UPROPERTY(EditAnywhere)
		double initDispZ = 0.0;
	UPROPERTY(EditAnywhere)
		TEnumAsByte<ForcingType> forcingType;
protected:
	virtual void Computation();
	virtual void Visualization(float deltaTime);
	virtual void BuildHUDLines();
	virtual void LogData();
	virtual void EndPlay(const EEndPlayReason::Type EndPlayReason) override;
	void Euler(double dT); 
	//(21) ---------
	void RKF(double dT);

	double floorPlateDispZ;
	//double drive = 0;
	FVector floorPlateStartLocation;
	FVector bodyStartLocation;
	FVector axleStartLocation;
	FVector Wheel_LF_StartLocation;
	FVector Wheel_RF_StartLocation;
	FVector Wheel_LR_StartLocation;
	FVector Wheel_RR_StartLocation;
	FVector Spring_LF_StartLocation;
	FVector Spring_LR_StartLocation;
	FVector Spring_RF_StartLocation;
	FVector Spring_RR_StartLocation;

	double maxUO = 0;
	double maxAccelZ = 0;

	// (19) -------------------
private:
	TArray<FDataRecord> dataRecords;

};
