// ============================================================================
// CBP + FBO = MAS22 Simulation Framework
// 20-12-21
// 28-01-22
// MAS22_SkyMaster.h
// ============================================================================

#pragma once

#include "CoreMinimal.h"
#include "GameFramework/Actor.h"
#include "MAS22_Actor.h"
#include "MAS22_SkyMaster.generated.h"

UENUM()
enum ForcingType_SM {
	periodic_SM	    UMETA(DisplayName = "Periodic"),
	none_SM			UMETA(DisplayName = "None")
};

USTRUCT()
struct FDataRecord_SM {
	GENERATED_BODY()
		double time, theta, omega;
};

UCLASS()
class MAS22_SCIENCEPARK_API AMAS22_SkyMaster : public AMAS22_Actor {
	GENERATED_BODY()

	UPROPERTY(VisibleDefaultsOnly)
		class UCapsuleComponent* CapsuleCollider;
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* Base;
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* LeftCar;
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* RightCar;

public:	

	AMAS22_SkyMaster();
	virtual void ActorInitialize() override;
	virtual void SetInitialConditions() override;
	virtual void InitializeSolver() override;	
	virtual void ComputeDT() override;

	UPROPERTY(EditAnywhere)
		double mass = 1.0;
	UPROPERTY(EditAnywhere)
		double armLength = 0.2824;
	UPROPERTY(EditAnywhere)
		double gravity = 9.81;
	UPROPERTY(EditAnywhere)
		double damping = 0.1;
	UPROPERTY(EditAnywhere)
		double scaling = 1.0;
	UPROPERTY(EditAnywhere)
		double forcingAmplitude = 0.5;
	UPROPERTY(EditAnywhere)
		double forcingPeriod = 1.0;
	UPROPERTY(EditAnywhere)
		double initTheta = 0;
	UPROPERTY(EditAnywhere)
		TEnumAsByte<ForcingType_SM> forcingType;

protected:
	virtual void Computation();
	virtual void Visualization(float deltaTime);
	virtual void BuildHUDLines();
	virtual void LogData();
	virtual void EndPlay(const EEndPlayReason::Type EndPlayReason) override;
	void Euler(double dT);
	void RKF(double dT);

	double maxTheta;
	double theta;
	double omega;
	double forcing;

public:	
	// Called every frame
	//virtual void Tick(float DeltaTime) override;
private:
	TArray<FDataRecord_SM> dataRecords;

};
