// ============================================================================
// CBP + FBO = MAS22 Simulation Framework
// 20-12-21
//
// ============================================================================


#include "MAS22_Turbine.h"
#include "MAS22_Structs.h"
#include "MAS22_Actor.h"
#include "Components/CapsuleComponent.h"
#include "Components/StaticMeshComponent.h"

// Sets default values
AMAS22_Turbine::AMAS22_Turbine()
{
 	// Set this actor to call Tick() every frame.  You can turn this off to improve performance if you don't need it.
	PrimaryActorTick.bCanEverTick = false;

	
	// Main Collider
	CapsuleCollider = CreateDefaultSubobject<UCapsuleComponent>("CapsuleCollider");
	CapsuleCollider->SetupAttachment(GetRootComponent());
	CapsuleCollider->SetCapsuleHalfHeight(1000);
	CapsuleCollider->SetCapsuleRadius(100);


	// First get the mesh for the turbine tower. Dean did not do this in code.

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset1(TEXT("/Game/TurbineModels/WindTurb_Base.WindTurb_Base"));
	UStaticMesh* Asset1 = MeshAsset1.Object;

	TowerMesh = CreateDefaultSubobject<UStaticMeshComponent>("TowerMesh");
	TowerMesh->SetStaticMesh(Asset1);
	//TowerMesh->SetRelativeScale3D(FVector(.2f, .2f, .2f));
	TowerMesh->SetupAttachment(CapsuleCollider);

	// The parent of the blades is the TowerMesh. All component rotations and translations are relative to this.

	// First get the mesh
	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset(TEXT("/Game/TurbineModels/WindTurb_Blade.WindTurb_Blade"));
	UStaticMesh* Asset = MeshAsset.Object;

	Blade1Mesh = CreateDefaultSubobject<UStaticMeshComponent>("Blade1Mesh");
	Blade1Mesh->SetStaticMesh(Asset);
	Blade1Mesh->SetupAttachment(TowerMesh);
	//Blade1Mesh->SetRelativeScale3D(FVector(.2f, .2f, .2f));
	Blade1Mesh->SetRelativeLocation(FVector(0, 50, 4000));

	//static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset2(TEXT("/Game/TurbineModels/WindTurb_Blade.WindTurb_Blade"));
	//UStaticMesh* Asset2 = MeshAsset2.Object;

	Blade2Mesh = CreateDefaultSubobject<UStaticMeshComponent>("Blade2Mesh");
	Blade2Mesh->SetupAttachment(TowerMesh);
	Blade2Mesh->SetStaticMesh(Asset);
	//Blade2Mesh->SetRelativeScale3D(FVector(.2f, .2f, .2f));
	Blade2Mesh->SetRelativeLocation(FVector(0, 50, 4000));
	FRotator bladeRot = FRotator((120), 0, 0);
	Blade2Mesh->SetRelativeRotation(bladeRot);

	Blade3Mesh = CreateDefaultSubobject<UStaticMeshComponent>("Blade3Mesh");
	Blade3Mesh->SetupAttachment(TowerMesh);
	Blade3Mesh->SetStaticMesh(Asset);
	//Blade3Mesh->SetRelativeScale3D(FVector(.2f, .2f, .2f));
	Blade3Mesh->SetRelativeLocation(FVector(0, 50, 4000));
	bladeRot = FRotator((240), 0, 0);
	Blade3Mesh->SetRelativeRotation(bladeRot);

	

}

// Sends a log message to indicate that it has been detected
void AMAS22_Turbine::Message() {
	//UE_LOG(Log_MAS22, Error, TEXT("Hello"));
}

// Called when the game starts or when spawned
//void ATurbine::BeginPlay() {
void AMAS22_Turbine::ActorInitialize() {

	Super::ActorInitialize();

	SetInitialConditions();

	ReadCpLambdaFile();

	BuildHUDLines();

	region3MessageShown = false;
}

void AMAS22_Turbine::SetInitialConditions() {
	Super::SetInitialConditions();
	//set vars to inputted initial vals
	beta = region2Beta;
	omega = omegaInit;
	windV = windVInit;

	omegaRated = tsrOpt * vRated / R;
	deltaBeta = 0.0;
}

//populate the cpl array with data from external text document
void AMAS22_Turbine::ReadCpLambdaFile() {
	//import data from txt file
	FString importedString;
	FString file = FPaths::ProjectConfigDir();
	file.Append(TEXT("CpLambda.txt"));
	FFileHelper::LoadFileToString(importedString, *file);

	//split by lines and commas and add data to 2d array
	TArray<FString> parsedString;
	importedString.ParseIntoArray(parsedString, TEXT("\n"), true);
	for (int i = 0; i < parsedString.Num(); i++) {
		CplArray.Add(FCplArray());
		TArray<FString> parsedLine;
		FString line = parsedString[i];
		line.ParseIntoArray(parsedLine, TEXT(","), true);
		for (int j = 0; j < parsedLine.Num(); j++) CplArray[i].Add(FCString::Atod(*parsedLine[j]));
		// GEngine->AddOnScreenDebugMessage(-1, 40.0f, FColor::Yellow,
		// 		FString::Printf(TEXT("Added val: %f"), CplArray[i][j]));
	}
}

void AMAS22_Turbine::setRunning() {
	Super::SetRunning();
	totalRot = 0;
	
}

// This is called in the computational loop to update the values for the HUD.
// Will vary according to the model being simulated
void AMAS22_Turbine::BuildHUDLines() {
	HUDLine line;
	Super::BuildHUDLines();
	HUDLines.Empty(); // Clears array ready for dynamic update
	line.message = actorName;
	line.colour = FColor::Yellow;
	line.value = 3.254;
	HUDLines.Add(line);
	line.message = FString::Printf(TEXT("time = %f"), time);
	line.colour = FColor::Red;
	line.value = 3.254;
	HUDLines.Add(line);
	line.message = FString::Printf(TEXT("omega = %f"), omega);
	line.colour = FColor::Blue;
	line.value = 6.254;
	HUDLines.Add(line);
	line.message = FString::Printf(TEXT("wind speed = %f"), windV);
	line.colour = FColor::Blue;
	line.value = 6.254;
	HUDLines.Add(line);
	line.message = FString::Printf(TEXT("Power= %f"), power);
	line.colour = FColor::Blue;
	line.value = 6.254;
	HUDLines.Add(line);
	
	if (windType == stepped) {
		line.message = FString::Printf(TEXT("wind type is stepped"));
		line.colour = FColor::Blue;
		line.value = 6.254;
		HUDLines.Add(line);
	}
	if (windType == series) {
		line.message = FString::Printf(TEXT("wind type is series"));
		line.colour = FColor::Blue;
		line.value = 6.254;
		HUDLines.Add(line);
	}
	if (windType == steady) {
		line.message = FString::Printf(TEXT("wind type is steady"));
		line.colour = FColor::Blue;
		line.value = 6.254;
		HUDLines.Add(line);
	}

	bHUD_exists = true;

}

void AMAS22_Turbine::ComputeDT() {
	if (bRunning) {
		for (int i = 0; i < speedUp; i++) Computation();
		Visualization(dT * scaling);
		BuildHUDLines();
		LogData();
		itn++;
	}
	if (time >= stopTime) {
		if (DEBUG) GEngine->AddOnScreenDebugMessage(-1, 4.0f, FColor::Yellow,
			FString::Printf(TEXT(">>>> SIMULATION STOPPED <<<<<")));
		bRunning = false;
	}
}

void AMAS22_Turbine::LogData() {
	int m;
	m = (int(logInterval / dT));
	if (itn % m == 0) {
		snapshots.Add(FSnapshot{ time,windV,omega,Cp,power,tsr,beta });
	}

}

void AMAS22_Turbine::Computation() {

	// assume turbine is not starting from rest: omega = omegaInit in BeginPlay()
	//CODERS
	// restart turbine if stops
	if (omega < 0.000001) {
		omega = omegaInit;
		GEngine->AddOnScreenDebugMessage(1, 1.0f, FColor::Yellow,
			FString::Printf(TEXT(">>>>>>> Turbine Restart")));
	}

	// clamp windV to minimum;
	if (windV < 0.01) windV = 0.01;

	//step the wind if option selected
	//if (useSteppedWindSpeed) windV = SteppedWindSpeed();
	// MODS FOR CODERS
	if (windType == stepped) windV = SteppedWindSpeed();
	if (windType == series && itn % 100 == 0)  windV = 5.1 + CalcWindSpeed(averageWindSpeed);

	torqueB = 0.0;

	// set region ----------------------------------------------------------------
	prevRegion = region;

	if (windV < vCutIn) region = 1;
	else if (windV > vCutOut) region = 4;
	else if (windV > vCutIn && omega < omegaRated) {
		region = 2;
	}
	else {
		if (!region3MessageShown) {
			//GEngine->AddOnScreenDebugMessage(-1, 5.0f, FColor::Yellow,
			//	FString::Printf(TEXT(">>>>>>>>>>> Entering Region-3 <<<<<<<<<<<<<<<<<<")));
			region3MessageShown = true;
		}
		region = 3;
	}
	//MODS FOR CODERS
	//if(omega > omegaRated)
	//	GEngine->AddOnScreenDebugMessage(-1, 2.0f, FColor::Yellow,
	//		FString::Printf(TEXT(">>>>>>>>>>> omega > omegaRated")));

	//R=20;
	//MODS FOR CODERS
	if (region != prevRegion)
		GEngine->AddOnScreenDebugMessage(1, 2.0f, FColor::Yellow,
			FString::Printf(TEXT(">>>> Transit from Region %d to region %d <<<<<<<<<<<"), prevRegion, region));

	// FSM run according to active region-----------------------------------------
	switch (region) {
	case 1:
		//add
		break;
	case 2:
		Region2();
		// MODS FOR CODERS
		// Make sure beta returns to Region2 value
		beta += 10 * (region2Beta - beta) * dT;
		break;
	case 3:
		deltaBeta = Kp * (omega - omegaRated) * dT;
		beta += deltaBeta;
		Region2();
		break;
	case 4:
		// This is a brutal dead stop, need to have braking effect here so omega
		// and power go gracefully to zero
		// CODERS
		omega -= omega * dT / 5.0;  //Time constant 5 sec
		//omega = 0.0;
		//power = 0.0;
		//Region2();
		break;
	}

	time += dT;
}

double AMAS22_Turbine::SteppedWindSpeed() {
	double newWindV;
	//GEngine->AddOnScreenDebugMessage(-1, 40.0f, FColor::Yellow,
	//	FString::Printf(TEXT("SteppedWindSpeed() stepTime = %e"), stepTime));
	if (stepTime >= windStepTimeInterval) {
		stepTime = 0;
		newWindV = windV + windStepHeight;
		//GEngine->AddOnScreenDebugMessage(-1, 40.0f, FColor::Yellow,
		//	FString::Printf(TEXT("time:%e,  windV=%e"),time,W));
		GEngine->AddOnScreenDebugMessage(-1, 4.0f, FColor::Yellow,
			FString::Printf(TEXT("SteppedWindSpeed() time:  %e,   Cp:  %e, beta = %e   windV:  %e"), time, Cp, beta, newWindV));
		return newWindV;
	}

	stepTime += dT;
	return windV;
}

//standard region 2 operation
void AMAS22_Turbine::Region2() {
	tsr = omega * R / windV;
	Cp = LookUpInCpLambdaArray(tsr, beta);
	//GEngine->AddOnScreenDebugMessage(-1, 40.0f, FColor::Yellow,
	//	FString::Printf(TEXT("time:%e   Cp:%e"), time,Cp));

	torqueW = (0.5 * rho * pi * (PowerOf(R, 5)) * (Cp / PowerOf(tsr, 3)) * PowerOf(omega, 2));
	if (tsr == 0) torqueW = 0;

	torqueG = (0.5 * rho * pi * (PowerOf(R, 5)) * (cpMax / PowerOf(tsrOpt, 3)) * PowerOf(omega, 2));

	error = (torqueW - torqueG) - torqueB;
	omega += error * dT / J;
	if (omega < 0) omega = 0;
	theta += omega * dT;
	power = torqueG * omega;

	//normalise
	if (theta > pi) theta = theta - 2 * pi;
	if (theta < -pi) theta = theta + 2 * pi;
	//if (theta > 2*pi) theta = theta - 2 * pi;
	//if (theta < -2*pi) theta = theta + 2 * pi;
}

double AMAS22_Turbine::Interpolate(double beta_, double beta1_, double beta2_, double Cp1_, double Cp2_) {
	double ret;
	ret = (Cp2_ - Cp1_) / (beta2_ - beta1_) * (beta_ - beta1_) + Cp1_;
	return ret;
}

double AMAS22_Turbine::LookUpInCpLambdaArray(double tsr_, double pitch_) {
	int row, col;
	double b1, b2, Cp1, Cp2;
	double Cpp;

	row = (int)((tsr_ - 3.0) * 10);
	if (row > 90) row = 90;
	if (row < 0) row = 0;

	// catch the limits of the pitch_

	if (pitch_ < 3.2) col = 0;
	if (pitch_ > 33.7) col = 12;

	// now do the interpolations.

	b1 = 3.2;
	b2 = 3.7;
	if ((pitch_ > b1) && (pitch_ <= b2)) {
		col = 0;
		Cp1 = CplArray[row][col];
		col = 1;
		Cp2 = CplArray[row][col];
		Cpp = Interpolate(pitch_, b1, b2, Cp1, Cp2);
		return Cpp;
	}

	b1 = 3.7;
	b2 = 4.2;
	if ((pitch_ > b1) && (pitch_ <= b2)) {
		col = 1;
		Cp1 = CplArray[row][col];
		col = 2;
		Cp2 = CplArray[row][col];
		Cpp = Interpolate(pitch_, b1, b2, Cp1, Cp2);
		return Cpp;
	}

	b1 = 4.2;
	b2 = 4.7;
	if ((pitch_ > b1) && (pitch_ <= b2)) {
		col = 2;
		Cp1 = CplArray[row][col];
		col = 3;
		Cp2 = CplArray[row][col];
		Cpp = Interpolate(pitch_, b1, b2, Cp1, Cp2);
		return Cpp;
	}

	b1 = 4.7;
	b2 = 5.7;
	if ((pitch_ > b1) && (pitch_ <= b2)) {
		col = 3;
		Cp1 = CplArray[row][col];
		col = 4;
		Cp2 = CplArray[row][col];
		Cpp = Interpolate(pitch_, b1, b2, Cp1, Cp2);
		return Cpp;
	}

	b1 = 5.7;
	b2 = 6.7;
	if ((pitch_ > b1) && (pitch_ <= b2)) {
		col = 4;
		Cp1 = CplArray[row][col];
		col = 5;
		Cp2 = CplArray[row][col];
		Cpp = Interpolate(pitch_, b1, b2, Cp1, Cp2);
		return Cpp;
	}

	b1 = 6.7;
	b2 = 7.7;
	if ((pitch_ > b1) && (pitch_ <= b2)) {
		col = 5;
		Cp1 = CplArray[row][col];
		col = 6;
		Cp2 = CplArray[row][col];
		Cpp = Interpolate(pitch_, b1, b2, Cp1, Cp2);
		return Cpp;
	}

	b1 = 7.7;
	b2 = 8.7;
	if ((pitch_ > b1) && (pitch_ <= b2)) {
		col = 6;
		Cp1 = CplArray[row][col];
		col = 7;
		Cp2 = CplArray[row][col];
		Cp = Interpolate(pitch_, b1, b2, Cp1, Cp2);
		return Cp;
	}

	b1 = 8.7;
	b2 = 13.7;
	if ((pitch_ > b1) && (pitch_ <= b2)) {
		col = 7;
		Cp1 = CplArray[row][col];
		col = 8;
		Cp2 = CplArray[row][col];
		Cpp = Interpolate(pitch_, b1, b2, Cp1, Cp2);
		return Cpp;
	}

	b1 = 13.7;
	b2 = 18.7;
	if ((pitch_ > b1) && (pitch_ <= b2)) {
		col = 8;
		Cp1 = CplArray[row][col];
		col = 9;
		Cp2 = CplArray[row][col];
		Cpp = Interpolate(pitch_, b1, b2, Cp1, Cp2);
		return Cpp;
	}

	b1 = 18.7;
	b2 = 23.7;
	if ((pitch_ > b1) && (pitch_ <= b2)) {
		col = 9;
		Cp1 = CplArray[row][col];
		col = 10;
		Cp2 = CplArray[row][col];
		Cpp = Interpolate(pitch_, b1, b2, Cp1, Cp2);
		return Cpp;
	}

	b1 = 23.7;
	b2 = 28.7;
	if ((pitch_ > b1) && (pitch_ <= b2)) {
		col = 10;
		Cp1 = CplArray[row][col];
		col = 11;
		Cp2 = CplArray[row][col];
		Cpp = Interpolate(pitch_, b1, b2, Cp1, Cp2);
		return Cpp;
	}

	b1 = 28.7;
	b2 = 33.7;
	if ((pitch_ > b1) && (pitch_ <= b2)) {
		col = 11;
		Cp1 = CplArray[row][col];
		col = 12;
		Cp2 = CplArray[row][col];
		Cpp = Interpolate(pitch_, b1, b2, Cp1, Cp2);
		return Cpp;
	}
	return 1;
}

//calculate rayleigh wind speed
double AMAS22_Turbine::CalcWindSpeed(double averageSpeed) {
	double test;
	double calcSpeed;
	double c;

	c = 2 * averageSpeed / FMath::Sqrt(pi);
	test = FMath::RandRange(0.0f, 1.0f);
	calcSpeed = c * FMath::Sqrt(-1 * FMath::Loge(test));

	return calcSpeed;
}

//power of function, cause the built in math one uses floats
//(WARNING DOESNT WORK FOR POWERS EQUALING:  0, NEGATIVE AND NOT WHOLE NUMBERS)
double AMAS22_Turbine::PowerOf(double baseNum, int expNum) {
	switch (expNum) {
	case 1:
		return baseNum;
	default:
		double answerVal = baseNum;
		for (int i = 1; i < expNum; i++) answerVal = answerVal * baseNum;
		return answerVal;
	}
}

//==========================================================================================================================================
//	VISUALIZATION
//==========================================================================================================================================

void AMAS22_Turbine::Visualization(float deltaTime) {
	// Following removed, placed into header as a private. Static variables seem to be shared between instances.
	//static float totalRot = 0;

	totalRot += omega * deltaTime;

	Blade1Mesh->SetRelativeRotation(FRotator(totalRot, 0, 0).Quaternion());
	Blade2Mesh->SetRelativeRotation(FRotator(totalRot + 120, 0, 0).Quaternion());
	Blade3Mesh->SetRelativeRotation(FRotator(totalRot + 240, 0, 0).Quaternion());
}

void AMAS22_Turbine::EndPlay(const EEndPlayReason::Type EndPlayReason) {
	Super::EndPlay(EndPlayReason);

	//get the file path
	FString logFile = FPaths::ProjectDir();
	FString fileName = "OctaveFiles/GeneratedPlots.m";
	fileName.Append("_");
	fileName.Append(actorName);
	logFile.Append(fileName);

	//add recorded vals

	int ticksRan = snapshots.Num();

	GEngine->AddOnScreenDebugMessage(-1, 40.0f, FColor::Yellow,
		FString::Printf(TEXT(">>>> EndPlay: number of logs = %d"), ticksRan));
	//time
	FString stringToSave = "times = [";
	for (int i = 0; i < ticksRan; i++) stringToSave.Append(FString::Printf(TEXT("%e,"), snapshots[i].time));
	stringToSave.Append("];\n");
	//wind velocity
	stringToSave.Append("windVs = [");
	for (int i = 0; i < ticksRan; i++) stringToSave.Append(FString::Printf(TEXT("%e,"), snapshots[i].windV));
	stringToSave.Append("];\n");
	//rotation speed
	stringToSave.Append("omegas = [");
	for (int i = 0; i < ticksRan; i++) stringToSave.Append(FString::Printf(TEXT("%e,"), snapshots[i].omega));
	stringToSave.Append("];\n");
	//error
	stringToSave.Append("Cps = [");
	for (int i = 0; i < ticksRan; i++) stringToSave.Append(FString::Printf(TEXT("%e,"), snapshots[i].Cp));
	stringToSave.Append("];\n");
	//power
	stringToSave.Append("powers = [");
	for (int i = 0; i < ticksRan; i++) stringToSave.Append(FString::Printf(TEXT("%e,"), snapshots[i].power));
	stringToSave.Append("];\n");
	//tip speed ratio
	stringToSave.Append("tsrs = [");
	for (int i = 0; i < ticksRan; i++) stringToSave.Append(FString::Printf(TEXT("%e,"), snapshots[i].tsr));
	stringToSave.Append("];\n");
	//beta
	stringToSave.Append("betas = [");
	for (int i = 0; i < ticksRan; i++) stringToSave.Append(FString::Printf(TEXT("%e,"), snapshots[i].beta));
	stringToSave.Append("];\n");

	//plot graphs
	//windV / time
	// Mod 14-11-21 CBP autoscale of time axis
	stringToSave.Append("maxTimes = floor(max(times));\n");

	stringToSave.Append("subplot(3,2,1)\n");
	stringToSave.Append("plot(times(:),windVs(:));\n");
	stringToSave.Append("xlabel('Time');\n");
	stringToSave.Append("ylabel('windV');\n");
	//CODERS
	stringToSave.Append("line([0, max(times)], [11.7, 11.7]);\n");
	stringToSave.Append("axis([0,floor(max(times)),0,max(windVs*1.1)]);\n");
	//omega / time
	stringToSave.Append("subplot(3,2,2)\n");
	stringToSave.Append("plot(times(:),omegas(:));\n");
	stringToSave.Append("xlabel('Time');\n");
	stringToSave.Append("ylabel('omega');\n");
	// CODERS
	stringToSave.Append("line([0, max(times)], [3.86, 3.86]);\n");
	stringToSave.Append("axis([0,floor(max(times)),0,max(omegas*1.1)]);\n");

	//error / time
	stringToSave.Append("subplot(3,2,3)\n");
	stringToSave.Append("plot(times(:),Cps(:));\n");
	stringToSave.Append("xlabel('Time');\n");
	stringToSave.Append("ylabel('Cp');\n");
	//CODERS
	stringToSave.Append("line([0, max(times)], [0.463, 0.463]);\n");
	stringToSave.Append("axis([0,floor(max(times)),0,max(Cps*1.1)]);\n");
	//power / time
	stringToSave.Append("subplot(3,2,4)\n");
	stringToSave.Append("plot(times(:),powers(:));\n");
	stringToSave.Append("xlabel('Time');\n");
	stringToSave.Append("ylabel('power');\n");
	// CODERS
	stringToSave.Append("line([0, max(times)], [600000, 600000]);\n");
	stringToSave.Append("axis([0,floor(max(times)),0,max(powers*1.1)]);\n");
	//tip speed ratio / time
	stringToSave.Append("subplot(3,2,5)\n");
	stringToSave.Append("plot(times(:),tsrs(:));\n");
	stringToSave.Append("xlabel('Time');\n");
	stringToSave.Append("ylabel('ts ratio');\n");
	//CODERS
	//They should work out how to get tsrOpt value in the string, so we don't need to hard-code this.
	stringToSave.Append("line([0, max(times)], [6.6, 6.6]);\n");
	stringToSave.Append("axis([0,floor(max(times)),0,max(tsrs*1.1)]);\n");
	//beta / time
	stringToSave.Append("subplot(3,2,6)\n");
	stringToSave.Append("plot(times(:),betas(:));\n");
	stringToSave.Append("xlabel('Time');\n");
	stringToSave.Append("ylabel('beta');\n");
	stringToSave.Append("axis([0,floor(max(times)),0,max(betas*1.1)]);\n");

	//generate the octave file
	FFileHelper::SaveStringToFile(stringToSave, *logFile);

	GetWorldTimerManager().ClearTimer(TimerHandle);
}



